use std::sync::Once;

use axum::{
    body::Body,
    extract::State,
    http::{header, Response, StatusCode, Uri},
    response::IntoResponse,
    routing::get,
};
use axum_typed_routing::{route, TypedRouter};
use maud::{html, Markup};
use rust_embed::RustEmbed;

#[route(GET "/hello")]
async fn hello_world(State(_): State<String>) -> Markup {
    html! {
        h1 { "Hello, World!" }
    }
}

#[allow(unused)]
#[route(GET "/item/:id?amount&offset")]
async fn item_handler(
    id: u32,
    amount: Option<u32>,
    offset: Option<u32>,
    State(state): State<String>,
    // Json(json): Json<u32>,
) -> Markup {
    // todo!("handle request")

    html! {
        h1 { "Item" }
        p {
            (format!("{id:?} {amount:?} {offset:?} {state:?}"))
        }
    }
}

// We use a wildcard matcher ("/dist/*file") to match against everything
// within our defined assets directory. This is the directory on our Asset
// struct below, where folder = "examples/public/".
#[route(GET "/dist/*path")]
async fn static_handler(path: String, _: State<String>) -> impl IntoResponse {
    let path = path.trim_start_matches('/').to_string();

    StaticFile(path)
}

fn markup_404(uri: String) -> Markup {
    html! {
        h1 { "404" }
        p { (format!("{uri:?} Not Found")) }
    }
}

fn markup_405() -> Markup {
    html! {
        h1 { "404" }
        p { "Method not allowed!" }
    }
}

// Finally, we use a fallback route for anything that didn't match.
async fn handle_404(uri: Uri) -> impl IntoResponse {
    (StatusCode::NOT_FOUND, markup_404(format!("{uri:?}"))).into_response()
}

async fn handle_405() -> impl IntoResponse {
    (StatusCode::METHOD_NOT_ALLOWED, markup_405()).into_response()
}

#[derive(RustEmbed)]
#[folder = "public/"]
struct Asset;

pub struct StaticFile<T>(pub T);

impl<T> IntoResponse for StaticFile<T>
where
    T: Into<String>,
{
    fn into_response(self) -> Response<Body> {
        let path = self.0.into();

        match Asset::get(path.as_str()) {
            Some(content) => {
                let mime = mime_guess::from_path(path).first_or_octet_stream();

                ([(header::CONTENT_TYPE, mime.as_ref())], content.data).into_response()
            }
            None => (StatusCode::NOT_FOUND, markup_404(path)).into_response(),
        }
    }
}

pub fn initialize_logger() {
    static INIT: Once = Once::new();

    INIT.call_once(|| {
        let env_filter = tracing_subscriber::EnvFilter::builder()
            .with_default_directive(tracing::level_filters::LevelFilter::INFO.into())
            .from_env_lossy();

        tracing_subscriber::fmt()
            .pretty()
            .with_env_filter(env_filter)
            .init();
    });
}

#[tokio::main]
async fn main() {
    initialize_logger();

    // TODO pick free port/config
    let listener = tokio::net::TcpListener::bind("0.0.0.0:8000").await.unwrap();

    let ui_router = axum::Router::new()
        .typed_route(item_handler)
        .typed_route(hello_world);

    let router: axum::Router = axum::Router::new()
        .merge(ui_router.clone())
        .nest("/ui", ui_router)
        .typed_route(static_handler)
        .fallback_service(get(handle_404))
        .method_not_allowed_fallback(handle_405)
        .with_state("state".to_string());

    let _ = axum::serve(listener, router.into_make_service()).await;
}
